# -*- coding: utf-8 -*-
"""
    torchrecorder.nodes
    ~~~~~~~~~~~~~~

    Nodes of the execution graph

    :copyright: (c) 2019 by Gautham Venkatasubramanian.
    :license: see LICENSE for more details.
"""


class BaseNode(object):
    """Wrapper object to encapsulate recorded information.

    Attributes:
        fn  (object):       an `object` recorded by the `~torchrecorder.recorder.Recorder`
        name (str):         name of the `.fn`
        depth (int):        `int`, scope depth of `.fn`
        parent (object):    a `.fn` in whose scope the current `.fn` exists
    """

    def __init__(self, name="", fn=None, depth=-1, parent=None):
        self.fn = fn
        self.name = name
        self.depth = depth
        self.parent = parent

    def __str__(self):
        internals = [
            "depth={}".format(self.depth),
            "fn_type={}".format(type(self.fn)),
            "parent_type={}".format(type(self.parent)),
        ]
        return self.name + "(" + ",".join(internals) + ")"

    def __repr__(self):
        return self.__str__()


class TensorNode(BaseNode):
    """Node to encapsulate a `torch.Tensor`.

    Attributes:
        fn( `torch.Tensor` ):
        name (str):         name of the `.fn`
        depth (int):        `int`, scope depth of `.fn`
        parent (object):    a `.fn` in whose scope the current `.fn` exists
    """

    pass


class ParamNode(TensorNode):
    """Node to encapsulate a `torch.nn.Parameter`.

    Attributes:
        fn( `torch.nn.Parameter` ):
        name (str):         name of the `.fn`
        depth (int):        `int`, scope depth of `.fn`
        parent (object):    a `~torch.nn.Module` whose `~torch.nn.Module.parameters` contains `.fn`
    """

    pass


class OpNode(BaseNode):
    """Node to encapsulate an Op, a ``grad_fn`` attribute
    of a `torch.Tensor`.

    Attributes:
        fn( `torch.Tensor` ):
        name (str):         name of the `.fn`
        depth (int):        `int`, scope depth of `.fn`
        parent (object):    a `~torch.nn.Module` in whose ``forward`` the current `OpNode.fn` was executed
    """

    pass


class LayerNode(BaseNode):
    """Node to encapsulate a `torch.nn.Module`.

    Attributes:
        fn( `torch.nn.Module` ):
        name (str):         name of the `.fn`
        depth (int):        `int`, scope depth of `.fn`
        parent (object):    a `~torch.nn.Module` in whose `~torch.nn.Module.forward`
                            `.fn` was called
        subnets (set):      a set `~torch.nn.Module` s or ``grad_fn`` s which are
                            called in `.fn` 's `~torch.nn.Module.forward`
        pre :     ``handle`` to the prehook on `.fn`
        post :    ``handle`` to the hook on `.fn`
    """

    def __init__(self, name="", fn=None, depth=-1, parent=None):
        BaseNode.__init__(self=self, name=name, fn=fn, depth=depth, parent=parent)
        self.pre = None
        self.post = None
        self.back = None
        self.subnets = set()

    def __str__(self):
        internals = [
            "depth={}".format(self.depth),
            "fn_type={}".format(type(self.fn)),
            "parent_type={}".format(type(self.parent)),
            "subnets_in_scope={}".format([type(x) for x in self.subnets]),
        ]
        return self.name + "(" + ",".join(internals) + ")"



# -*- coding: utf-8 -*-
"""
    torchrecorder.renderer.base
    ~~~~~~~~~~~~~~~~~~~~~~

    Abstract base class for Renderer objects

    :param copyright: (c) 2020 by Gautham Venkatasubramanian.
    :param license: see LICENSE for more details.
"""
from collections import OrderedDict
from ..nodes import LayerNode


class BaseRenderer(object):

    """Base Class for rendering information from a `~torchrecorder.recorder.Recorder`.

    Attributes:

        rec (`~torchrecorder.recorder.Recorder`):
        render_depth (int): nodes having a greater depth than this value
                            will not be rendered
        processed (`collections.OrderedDict`):
                            An ``OrderedDict`` whose keys contain ``nodes`` and values
                            contain the corresponding (directed) edge lists
    """

    def __init__(self, rec, render_depth=256):
        self.rec = rec
        self.render_depth = render_depth
        self.processed = OrderedDict()

    def render_node(self, dest, node):
        raise NotImplementedError("Base Class")

    def render_recursive_node(self, dest, node):
        raise NotImplementedError("Base Class")

    def render_edge(self, dest, fnode, tnode):
        raise NotImplementedError("Base Class")

    def __call__(self, dest):
        """Render nodes and edges.

        Uses `.render_depth` to select nodes and edges from `.rec`
        required for rendering. calls `.render_node` on each node,
        followed by `render_edge` on each edge from that node.

        Args:
            dest:    destination for the rendered information
                    (`graphviz.Digraph`, `dict` etc.)
        Returns:
            ``dest`` after updating with necessary information
        """
        self.processed.clear()
        self._process_nodes()
        self._process_edges()
        while len(self.processed) != 0:
            node = next(iter(self.processed))
            targets = self.processed[node]
            self.render_node(dest, node)
            for t in targets:
                self.render_edge(dest, node, t)
            self.processed.pop(node)
        return dest

    def _process_nodes(self):
        """Filter out nodes that have a greater depth than required.
        """
        for k, v in self.rec.nodes.items():
            if k is not None and v.depth <= self.render_depth:
                self.processed[v] = []

    def _process_edges(self):
        """Construct necessary edges between filtered nodes.

        After all nodes deeper than `.render_depth` have been removed, all the
        edges that between these nodes and those that remain must be
        transformed accordingly: such edges are "lifted up" for rendering, and
        ignored if they are internal to a node (i.e. both source and
        destination have been removed).
        """

        def lifted_node(x):
            xnode = self.rec.nodes[x]
            while xnode.depth > self.render_depth:
                xnode = self.rec.nodes[xnode.parent]
            return xnode

        lifted_edges = set(
            (lifted_node(x), lifted_node(y), z)
            for x, y, z in self.rec.edges
            if lifted_node(x) != lifted_node(y)
        )
        for fnode, tnode, _ in lifted_edges:
            self.processed[fnode].append(tnode)



# -*- coding: utf-8 -*-
"""
    torchrecorder.renderer.__init__
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyright: (c) 2020 by Gautham Venkatasubramanian.
    :license: see LICENSE for more details.
"""
from .gv import GraphvizRenderer, GraphvizStyler



# -*- coding: utf-8 -*-
"""
    torchrecorder.recorder
    ~~~~~~~~~~~~~~~~~

    Uses hooks to record the traversal of the execution graph

    :copyright: (c) 2019 by Gautham Venkatasubramanian.
    :license: see LICENSE for more details.
"""
from torch.nn import Module
from collections import OrderedDict
from .nodes import BaseNode, TensorNode, ParamNode, OpNode, LayerNode
from functools import partial
import time


class Recorder(object):

    """Record and store execution graph information

    Attributes:
        fn_set (set):         a set of objects ( `~torchrecorder.nodes.BaseNode.fn`\ s) that contain recordable information
        nodes (dict):           a mapping of `~torchrecorder.nodes.BaseNode.fn`\ s
                                to their corresponding `~torchrecorder.nodes.BaseNode`\ s
        fn_types (dict):      a count of `~torchrecorder.nodes.BaseNode.fn`\ s by type for naming
        edges   (set(tuple)):   a set of edges, each a pair of `~torchrecorder.nodes.BaseNode.fn`\ s
    """

    def __init__(self):
        self.nodes = OrderedDict()
        self.fn_types = dict()
        self.fn_set = set()
        self.edges = set()

        self._start_time = None
        self._create_context()

    def add_node(self, net, depth=0, parent=None, name=None):
        """Construct a node of recording graph.

        Construct a `~.nodes.BaseNode` that will store information related to ``net``
        as the neural network is run.

        Args:

            net :       Object whose information will be stored as the ``fn``
                        attribute of a `~.nodes.BaseNode`
            depth :     The scope depth at which ``net`` is found
            parent :    The object as part of which net will be run
            name :      a name to recognize the object during rendering, defaults to class name

        Returns:
            `None`

        """
        classname = type(net).__name__
        if self.fn_types.get(classname):
            self.fn_types[classname] += 1
        else:
            self.fn_types[classname] = 1

        if name is None:
            objname = classname
            if self.fn_types[classname] > 1:
                objname = objname + "-" + str(self.fn_types[classname])
        else:
            objname = name

        if isinstance(net, Module):
            if depth > 0 and name is not None:
                objname = objname + "\n(" + classname + ")"
            x = LayerNode(name=objname, fn=net, parent=parent, depth=depth)

            x.pre = net.register_forward_pre_hook(partial(prehook, rec=self, node=x))
            x.post = net.register_forward_hook(partial(posthook, rec=self, node=x))
            # x.back = net.register_backward_hook(partial(backhook, rec=self, node=x))
        elif "Tensor" in classname:
            x = TensorNode(name=objname, fn=net, parent=parent, depth=depth)
        elif "Parameter" in classname:
            x = ParamNode(name=objname, fn=net, parent=parent, depth=depth)
        elif hasattr(net, "next_functions"):
            x = OpNode(name=objname, fn=net, parent=parent, depth=depth)
        else:
            raise RuntimeError("Cannot create node for " + str(net))

        self.nodes[net] = x
        self.fn_set.add(net)
        if x.parent is not None:
            pnode = self.nodes[x.parent]
            pnode.subnets.add(net)

    def add_dummy(self, dummy, fn):
        """Point to an existing node to assist recording.

        Instead of creating a separate node, the ``dummy`` object is used to point
        to an existing node containing ``fn``. Used for dummy ops and
        ``AccumulateGradient``\ s (see `leaf_dummy` ).

        Args:
            dummy: a dummy `torch.Tensor` or op that should not be recorded
            fn : a recorded object that will be connected to further ops

        """
        self.fn_set.add(dummy)
        self.nodes[dummy] = self.nodes[fn]

    def add_edge(self, _from, _to):
        """Construct an edge of the recording graph.

        Records an edge between two `~torchrecorder.nodes.BaseNode.fn` objects to be used while rendering.
        This will be used along with the ``nodes`` dictionary to map edges properly.

        Args:
            _from (`~torchrecorder.nodes.BaseNode.fn`\ ):
            _to (`~torchrecorder.nodes.BaseNode.fn`\ ):
        """
        if _from is None or _to is None:
            raise AssertionError("Cannot draw edge involving" + str((_from, _to)))
        if self._start_time is not None:
            timestamp = time.time() - self._start_time
        else:
            timestamp = 0
            self._start_time = time.time()
        edge = (_from, _to, round(timestamp, 6))
        self.edges.add(edge)

    def register_hooks(self, net, depth=0, parent=None, name=None):
        """Register the hooks of the `.Recorder` recursively on
        a `torch.nn.Module`\ .

        The hooks registered are `~functools.partial` versions
        of `prehook` and `posthook` corresponding to each node.

        Args:
            net (`~torch.nn.Module`\ ):
            depth (int):
            parent (`torch.nn.Module`\ ): the parent of ``net``
            name (str): name of ``net``

        Returns:
            `None`

        """
        self.add_node(net, depth, parent, name)
        for n, x in net.named_children():
            self.register_hooks(x, depth=depth + 1, parent=net, name=n)

    def remove_hooks(self):
        """Remove hooks from any `~torch.nn.Module`\ s in
        `~torchrecorder.nodes.LayerNode`\ s.

        After the recording is completed, the hooks in
        `~torchrecorder.nodes.LayerNode`\ s are unnecessary.
        They are removed to prevent any possible issues.
        """
        for node in set(self.nodes.values()):
            if isinstance(node, LayerNode):
                node.pre.remove()
                node.post.remove()

    def _create_context(self):
        """Construct a dummy node as the context for the recording graph.

        Adds a dummy node (mapped to `None`) to the recording graph as the
        context.  This node has the least depth (-1); is its own parent, and
        will have only one child: the network whose execution is to be
        recorded.

        Returns:
            `None`
        """
        self.fn_set.add(None)
        self.nodes[None] = BaseNode(fn=None, depth=-1, parent=None, name="ContextDummy")


def op_acc(gf, rec, node):
    """Operator Accumulator.

    Creates an `~.nodes.OpNode` to record the newly-performed operation ``gf``, if not
    already recorded. If ``gf`` is an initialization op (``AccumulateGradient``),
    then points ``gf`` to its connected `torch.Tensor` instead of creating an
    `~.nodes.OpNode`. Otherwise recursively checks all operations that are connected to
    ``gf`` and adds them if necessary.

    Args:
        gf:    current operation, a ``grad_fn`` object obtained from a `torch.Tensor`
        rec:   a `~.Recorder` object whose nodes are updated
        node:  `~.nodes.LayerNode` whose ``fn`` the current operation is a part of

    Returns:
        `None`

    """
    if gf in rec.fn_set:
        pass
    else:
        if hasattr(gf, "variable"):
            rec.add_dummy(dummy=gf, fn=gf.variable)
        elif hasattr(gf, "next_functions"):
            rec.add_node(gf, node.depth + 1, node.fn)
            for x, y in gf.next_functions:
                if x is not None:
                    op_acc(x, rec, node)
                    rec.add_edge(_from=x, _to=gf)


def tensor_acc(tensor, rec, node):
    """Tensor Accumulator.

    Creates a `~.nodes.TensorNode` to record the newly-created tensor, if not already
    recorded.  Note that the resulting `~.nodes.TensorNode` has the same parent as
    ``node``, because the ``tensor`` is the output of/input to ``node.fn``.

    Args:
        tensor:    a `torch.Tensor`
        rec:       a `~.Recorder` object whose nodes are updated
        node:      a `~.nodes.LayerNode` whose ``fn`` outputs/inputs ``tensor``

    Returns:
        `None`

    """
    if tensor in rec.fn_set:
        pass
    else:
        rec.add_node(tensor, depth=node.depth, parent=node.parent)


def param_acc(param, rec, node):
    """Parameter Accumulator.

    Creates a `~.nodes.ParamNode` to record the parameter ``param`` of ``node.fn``, if not
    already recorded.  Note that ``node.fn`` is the *parent* of ``param``\ .

    Args:
        param:     a `~torch.nn.Parameter`
        rec:       the `~.Recorder` object whose nodes are updated
        node:      `~.nodes.LayerNode` whose ``fn`` contains ``param``

    Returns:
	`None`

    """
    if param in rec.fn_set:
        pass
    else:
        rec.add_node(param, depth=node.depth + 1, parent=node.fn)


def leaf_dummy(tensor, rec):
    """Performs a dummy operation (adding 0) to a leaf `~torch.Tensor`.

    This ensures that the (possibly in-place) operations performed on
    ``tensor`` hereafter can be correctly mapped. The dummy tensor (and
    operation) are not recorded separately, they merely point to the original
    tensor.

    Args:
        tensor:    a newly-formed leaf `torch.Tensor`
        rec:       the `~.Recorder` object whose nodes are updated

    Returns:
        ``tensor`` after adding 0

    """
    dummy = tensor + 0
    rec.add_dummy(dummy=dummy, fn=tensor)
    rec.add_dummy(dummy=dummy.grad_fn, fn=tensor)
    return dummy


def prehook(module, inputs, rec, node):
    """hook to record BEFORE the given ``module`` is run.

    Records parameters contained in ``module``, then checks each tensor in
    ``inputs`` for any operations that may have run after the end of the
    previous ``module``. The ``inputs`` are then converted to leaf tensors and
    recorded before being passed off to the ``module``.

    Args:
        module:     a `torch.nn.Module`
        inputs:     a `torch.Tensor` or a `tuple` of `torch.Tensor`\ s
        rec:        a `~.Recorder` object for global information
        node (`~torchrecorder.nodes.LayerNode`\ ):
                    ``node.fn`` is ``module``\ .

    Returns:
        the ``leaf``-equivalent of ``inputs``.

    """
    for name, param in module.named_parameters(recurse=False):
        param_acc(param, rec, node)
        if name is not None and name != "":
            rec.nodes[param].name = name
    is_singleton = not isinstance(inputs, tuple)
    a = [inputs] if is_singleton else inputs  # same input appearing multiple times?
    new_inputs = []
    for x in a:
        gf = x.grad_fn
        op_acc(gf, rec, rec.nodes[node.parent])
        tensor_acc(x, rec, node)
        if gf is not None:
            rec.add_edge(_from=gf, _to=x)
        new_inputs.append(leaf_dummy(x, rec))
    return new_inputs[0] if is_singleton else tuple(new_inputs)


def posthook(module, inputs, outputs, rec, node):
    """hook to record AFTER the given ``module`` has run and returned.

    Records any operations that may have run as part of ``module``\ , then
    checks if each tensor in the ``outputs`` has already been recorded by a
    sub\ ``module`` of the current ``module`` (the sub\ ``module``\ 's
    `posthook` would execute first!). If necessary, the ``outputs`` are
    converted to leaf tensors to record operations afresh.

    Args:
        module:     a `torch.nn.Module`
        inputs:     a `torch.Tensor` or a tuple of `torch.Tensor`\ s
        outputs:    a `torch.Tensor` or a tuple of `torch.Tensor`\ s
        rec:        a `~.Recorder` object for global information
        node (`~torchrecorder.nodes.LayerNode`\ ):
                    ``node.fn`` is ``module``.

    Returns:
        the ``leaf``-equivalent of ``outputs``.

    """
    is_singleton = not isinstance(outputs, tuple)
    b = [outputs] if is_singleton else outputs
    new_outputs = []
    for x in b:
        gf = x.grad_fn
        if gf not in rec.fn_set:
            x = x.detach()
            x.requires_grad = True
            tensor_acc(x, rec, node)
            op_acc(gf, rec, node)
            rec.add_edge(gf, x)
            new_outputs.append(leaf_dummy(x, rec))
        else:
            # if the op has already been recorded
            # it has to be a dummy op
            rec.nodes[gf].parent = node.parent
            rec.nodes[gf].depth -= 1
            node.subnets.remove(rec.nodes[gf].fn)
            new_outputs.append(x)
    return new_outputs[0] if is_singleton else tuple(new_outputs)


def backhook(module, grad_inputs, grad_outputs, rec, node):
    pass



# -*- coding: utf-8 -*-
"""
    torchrecorder.renderer.gv
    ~~~~~~~~~~~~~~~~~~~~

    Graphviz renderer object

    :param copyright: (c) 2020 by Gautham Venkatasubramanian.
    :param license: see LICENSE for more details.
"""
from ..nodes import BaseNode, TensorNode, ParamNode, OpNode, LayerNode
from .base import BaseRenderer
from graphviz import Digraph


class GraphvizStyler(object):
    """Provide styling options before rendering to graphviz.

    Attributes:
        styles (dict):   contains style properties for each subclass of `~torchrecorder.nodes.BaseNode`
    """

    def __init__(self, **styler_args):
        def setdefault(x, y):
            if styler_args.get(x, None) is None:
                styler_args[x] = y

        setdefault("style", "filled")
        setdefault("align", "left")
        setdefault("color", "black")

        self.styles = {
            BaseNode: dict(**styler_args),
            TensorNode: dict(**styler_args, fillcolor="lightblue"),
            ParamNode: dict(**styler_args, fillcolor="darkolivegreen"),
            OpNode: dict(**styler_args, fillcolor="orange", shape="box"),
            LayerNode: dict(**styler_args, fillcolor="lightgrey", shape="box"),
        }

    def style_node(self, node):
        """Construct style properties for the given node.

        Can be overridden to perform custom styling.

        Args:
            node (`~torchrecorder.nodes.BaseNode`\ ):
        Returns:
            a `dict` containing the required style properties

        """
        z = dict(**self.styles[type(node)])
        if isinstance(node, TensorNode):
            z["label"] = node.name + "\n" + str(list(node.fn.shape))
        else:
            z["label"] = node.name
        return z

    def style_edge(self, fnode, tnode):
        """Construct style properties to render the given edge

        Args:
            fnode: `~torchrecorder.nodes.BaseNode`
            tnode: `~torchrecorder.nodes.BaseNode`

        Returns:
            a `dict` containing the required style properties

        """
        return {}


class GraphvizRenderer(BaseRenderer):
    """Render information from a `~torchrecorder.recorder.Recorder` into a `graphviz.Digraph`.

    Attributes:
        styler (`class`): `.GraphvizStyler` or a subclass
    """

    def __init__(self, rec, render_depth=256, styler_cls=None, **styler_args):
        BaseRenderer.__init__(self, rec, render_depth)
        if styler_cls is None:
            styler_cls = GraphvizStyler
        self.styler = styler_cls(**styler_args)
        self.recursion_trace = []

    def render_node(self, g, node):
        """Render a node in `graphviz`

        Renders ``node`` into the `~graphviz.Digraph` ``g``,
        after applying appropriate styling.
        If ``node`` is a `~torchrecorder.nodes.LayerNode`, checks
        `.render_depth` to see if its
        `~.torchrecorder.nodes.LayerNode.subnets` have to rendered.

        Args:
            g (`graphviz.Digraph`):
            node (`~torchrecorder.nodes.BaseNode`):

        """
        if isinstance(node, LayerNode) and node.depth < self.render_depth:
            self.recursion_trace.append(g)
            self.render_recursive_node(g, node)
            self.recursion_trace.remove(g)
        else:
            style = self.styler.style_node(node)
            g.node(name=str(id(node)), **style)

    def render_recursive_node(self, g, node):
        """Render a `~torchrecorder.nodes.LayerNode` and its subnets.

        Args:
            g (`graphviz.Digraph`):
            node (`~torchrecorder.nodes.LayerNode`): has a
                                `~torchrecorder.nodes.LayerNode.depth` greater than
                                `.render_depth`

        The ``node`` is rendered as a separate `~graphviz.Digraph`
        and then is added as a `graphviz.Digraph.subgraph` to ``g``.
        """
        subg_style = self.styler.style_node(node)
        subg_style["fillcolor"] = "white"
        subg = Digraph(
            name="cluster_" + str(id(node)),
            graph_attr=subg_style,
            node_attr={"group": str(node.depth)},
        )
        for s in node.subnets:
            fnode = self.rec.nodes[s]
            self.render_node(subg, fnode)
        for s in node.subnets:
            fnode = self.rec.nodes[s]
            for tnode in self.processed[fnode]:
                if fnode.depth == tnode.depth:
                    self.render_edge(subg, fnode, tnode)
                else:
                    depth_diff = abs(fnode.depth - tnode.depth)
                    self.render_edge(self.recursion_trace[-depth_diff], fnode, tnode)
            self.processed.pop(fnode)
        g.subgraph(subg)

    def render_edge(self, g, fnode, tnode):
        """Render an edge in `graphviz`

        Args:
            g (`graphviz.Digraph`):
            fnode (`~torchrecorder.nodes.BaseNode`):
            tnode (`~torchrecorder.nodes.BaseNode`):

        """
        style = self.styler.style_edge(fnode, tnode)
        g.edge(str(id(fnode)), str(id(tnode)), **style)



from setuptools import setup, find_packages

VERSION = "1.0.2"

with open("./README.md") as f:
    long_description = f.read()

setup_info = dict(
    name="torchrecorder",
    version=VERSION,
    author="Gautham Venkatasubramanian",
    author_email="ahgamut@gmail.com",
    url="https://github.com/ahgamut/torchrecorder",
    description="Record execution graphs of PyTorch neural networks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    packages=find_packages("src"),
    package_dir={"": "src"},
    zip_safe=True,
    install_requires=["torch>=1.3", "graphviz"],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering :: Visualization",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
    ],
)

setup(**setup_info)



# -*- coding: utf-8 -*-
"""
    torchrecorder.helpers
    ~~~~~~~~~~~~~~~~

    Helper functions for torchrecorder

    :copyright: (c) 2020 by Gautham Venkatasubramanian.
    :license: see LICENSE for more details.
"""
from torch import randn
from .recorder import Recorder
from .renderer.gv import GraphvizRenderer
from graphviz import Digraph


def render_network(
    net,
    name,
    input_shapes,
    directory,
    filename=None,
    fmt="svg",
    input_data=None,
    render_depth=1,
    **styler_args
):
    """Render the structure of a `torch.nn.Module` to an image via `graphviz`.

    Args:
        net (`torch.nn.Module`):
        name (str): name of the network
        input_shapes (None, tuple or list(tuple)):
                    `tuple` if ``net`` has a single input,
                    `list` ( `tuple` ), `None`
                    if ``input_data`` is provided
        directory (str): directory to store the rendered image
        fmt (str, optional): image format
        input_data (`torch.Tensor` or `tuple` (`torch.Tensor` ), optional):
                    if ``net`` requires normalized inputs,
                    provide them here instead of setting ``input_shapes``.
        render_depth (int, optional): Default ``1``.
        **styler_args : node attributes to pass to `graphviz`

    """
    net = net.cpu().train()
    rec = record(net, name, input_shapes, input_data)
    g = make_dot(rec, render_depth, styler_cls=None, **styler_args)
    g.format = fmt
    g.attr(label="{} at depth = {}".format(name, render_depth))
    g.render("{}-{}".format(name, render_depth), directory=directory, cleanup=True)


def record(net, name, input_shapes, input_data=None):
    """Record the graph by running a single pass of a `torch.nn.Module`.

    Args:
        net (`torch.nn.Module`):
        name (str): name of the network
        input_shapes (None, tuple or list(tuple)):
                    `tuple` if ``net`` has a single input,
                    `list` ( `tuple` ), `None`
                    if ``input_data`` is provided
        input_data (`torch.Tensor` or `tuple` (`torch.Tensor` ), optional):
                    if ``net`` requires normalized inputs,
                    provide them here instead of setting ``input_shapes``.

    Returns:
        a `~.Recorder` object containing the execution graph

    """
    rec = Recorder()
    rec.register_hooks(net, depth=0, parent=None, name=name)

    data = []
    data_given = input_data is not None
    single_input = False
    if data_given:
        data = input_data
        single_input = not isinstance(input_data, tuple)
    else:
        if isinstance(input_shapes, list):
            single_input = False
            for shape in input_shapes:
                d = randn(shape)
                data.append(d)
            data = tuple(data)
        elif isinstance(input_shapes, tuple):
            single_input = True
            data = randn(input_shapes)

    if single_input:
        data.requires_grad = True
        rec.add_node(data, depth=0, parent=None, name="Input")
        pred = net(data)
    else:
        for i, d in enumerate(data):
            d.requires_grad = True
            rec.add_node(d, depth=0, parent=None, name="Input-{i}".format(i=i + 1))
        pred = net(*data)

    single_output = not isinstance(pred, tuple)
    if single_output:
        rec.nodes[pred].name = "Output"
    else:
        for i, p in enumerate(pred):
            rec.nodes[p].name = "Output-{i}".format(i=i + 1)

    rec.remove_hooks()
    return rec


def make_dot(rec, render_depth=256, styler_cls=None, **styler_args):
    """ Produces Graphviz representation from a `~torchrecorder.recorder.Recorder` object

    Args:
        rec (`~torchrecorder.recorder.Recorder`\ ):
        render_depth (int):     depth until which nodes should be rendered
        styler_cls:             styler class to instantiate when styling nodes.
                                If `None`, defaults to `.GraphvizStyler`.
    Kwargs:
        styler_args (optional): styler properties to be set for all nodes

    Returns:
        a `graphviz.Digraph` with the rendered nodes

    """
    graph_attr = dict(compound="true", ranksep="0.5", fontsize="24")
    node_attr = dict(fontsize="20")
    if styler_args.get("fontname", None) is not None:
        graph_attr["fontname"] = styler_args.get("fontname")
        node_attr["fontname"] = styler_args.get("fontname")
    g = Digraph(graph_attr=graph_attr, node_attr=node_attr)
    renderer = GraphvizRenderer(
        rec=rec, render_depth=render_depth, styler_cls=styler_cls, **styler_args
    )
    return renderer(g)
